#include "ParGraph.hpp"
#include <iostream>
#include <stdlib.h>

namespace ParGraph {
comm_layer_t _libbase;  // Die globalen Daten der Library
                        // insbesondere MPI-Kommunikator

inline void ParGraph_common_init()
{
    int tmp;

    MPI_Comm_rank(_libbase._comm,&tmp);
    _libbase._rank = size_t(tmp);
    MPI_Comm_size(_libbase._comm,&tmp);
    _libbase._size = size_t(tmp);
}

void ParGraph_init( int& argc, char**& argv )
{
    int mpi_is_init;
    MPI_Initialized(&mpi_is_init);

    _libbase._mpi_was_init = mpi_is_init;

    if ( !mpi_is_init )
        MPI_Init(&argc,&argv);

    MPI_Comm_dup(MPI_COMM_WORLD, &_libbase._comm);

    ParGraph_common_init();
}

void ParGraph_init(int& argc, char**& argv, MPI_Comm& user_comm)
{
    _libbase._mpi_was_init = 1;

    MPI_Comm_dup(user_comm, &_libbase._comm);

    ParGraph_common_init();
}

void ParGraph_finalize()
{
    MPI_Barrier(_libbase._comm);
    MPI_Comm_free(&_libbase._comm);

    if ( !_libbase._mpi_was_init )
    {
        MPI_Finalize();
    }
}

bool PG_ReduceOr( bool b )
{
  int val = b ? -1 : 0;
  int res;

  MPI_Allreduce( &val, &res, 1, MPI_INT, MPI_LOR, _libbase.comm() );

  return res != 0 ? true : false;
}

bool PG_ReduceAnd( bool b )
{
  int val = b ? -1 : 0;
  int res;

  MPI_Allreduce( &val, &res, 1, MPI_INT, MPI_LAND, _libbase.comm() );

  return res != 0 ? true : false;
}

size_t PG_ReduceAdd( size_t local )
{
  size_t global;
  MPI_Allreduce( &local, &global, 1, MPI_UNSIGNED, MPI_SUM, _libbase.comm() );
  return global;
}

AllGatherBuffer::AllGatherBuffer()
{
  recv_buffer_parts = new int[ _libbase.size() ];
  recv_size = new int[ _libbase.size() ];

  recv_buffer = 0;
  cur_recv_position = 0;
  whole_recv_size = 0;

  send_buffer = 0;
  send_size = 0;
  send_maxsize = 0;
  send_position = 0;
}

AllGatherBuffer::~AllGatherBuffer()
{
  if( send_buffer != 0 ) delete[] send_buffer;
  if( recv_buffer != 0 ) delete[] recv_buffer;
  delete[] recv_buffer_parts;
  delete[] recv_size;
}

void
AllGatherBuffer::communicate()
{
  if( recv_buffer != 0 ) delete[] recv_buffer;

  MPI_Allgather( &send_size, 1, MPI_INT, recv_size, 1, MPI_INT, _libbase.comm() );

  size_t size = 0;
  for( unsigned int i = 0 ; i < _libbase.size() ; i++ )
  {
    recv_buffer_parts[i] = size;
    size += recv_size[i];
  }
  recv_buffer = new char[ size ];
  cur_recv_position = 0;
  whole_recv_size = size;
  last_recv_proc = 0;

  MPI_Allgatherv( (int*) send_buffer, send_size, MPI_CHAR,
		 recv_buffer, recv_size, recv_buffer_parts, MPI_CHAR, _libbase.comm() );

  send_size = 0;
  send_maxsize = 0;
  send_position = 0;

  if( send_buffer != 0 )
  {
    delete[] send_buffer;
    send_buffer = 0;
  }
}

AllToAllBuffer::AllToAllBuffer()
{
  send_buffer = new char*[ _libbase.size() ];
  recv_buffer_parts = new int[ _libbase.size() ];
  send_size = new int[ _libbase.size() ];
  recv_size = new int[ _libbase.size() ];
  send_maxsize = new int[ _libbase.size() ];
  send_position = new int[ _libbase.size() ];

  recv_buffer = 0;
  cur_recv_position = 0;
  whole_recv_size = 0;

  for( unsigned int i = 0 ; i < _libbase.size() ; i++ )
  {
    send_buffer[i] = 0;
    send_size[i] = 0;
    send_maxsize[i] = 0;
    send_position[i] = 0;
  }
}

AllToAllBuffer::~AllToAllBuffer()
{
  for( unsigned int i = 0 ; i < _libbase.size() ; i++ )
  {
    free( send_buffer[ i ] );
  }
  delete[] send_buffer;
  delete[] send_size;
  delete[] send_maxsize;
  delete[] send_position;
  delete[] recv_buffer;
  delete[] recv_buffer_parts;
  delete[] recv_size;
}

void
AllToAllBuffer::communicate()
{
  if( recv_buffer != 0 ) delete[] recv_buffer;

  MPI_Alltoall( send_size, 1, MPI_INT, recv_size, 1, MPI_INT, _libbase.comm() );

  size_t size = 0;
  for( proc_t i = 0 ; i < _libbase.size() ; i++ )
  {
    recv_buffer_parts[i] = size;
    size += recv_size[i];
  }
  recv_buffer = new char[ size ];
  whole_recv_size = size;
  last_recv_proc = 0;
  cur_recv_position = 0;

  MPI_Alltoallv( 0, send_size, (int*) send_buffer, MPI_CHAR,  // WORKAROUND
		 recv_buffer, recv_size, recv_buffer_parts, MPI_CHAR, _libbase.comm() );

  for( proc_t i = 0 ; i < _libbase.size() ; i++ )
  {
    free( send_buffer[ i ] );
    send_buffer[ i ] = 0;
    send_size[i] = 0;
    send_maxsize[i] = 0;
    send_position[i] = 0;
  }
}
}  // namespace ParGraph
